@@ -33,7 +33,7 @@ import tempfile
 from contextlib import contextmanager
 from requests_toolbelt import MultipartEncoder
 
-__version__ = '2.2.1'
+__version__ = '2.3.0'
 __license__ = 'Apache 2.0'
 
 DEFAULT_PAGE_LIMIT = 50  # Number of results to fetch per request
@@ -323,7 +323,7 @@ class GirderClient(object):
             resp = self.post('api_key/token', parameters={
                 'key': apiKey
             })
-            self.token = resp['authToken']['token']
+            self.setToken(resp['authToken']['token'])
         else:
             if interactive:
                 if username is None:
@@ -343,7 +343,16 @@ class GirderClient(object):
             if 'authToken' not in resp:
                 raise AuthenticationError()
 
-            self.token = resp['authToken']['token']
+            self.setToken(resp['authToken']['token'])
+
+    def setToken(self, token):
+        """
+        Set a token on the GirderClient instance. This is useful in the case
+        where the client has already been given a valid token, such as a remote job.
+
+        :param token: A string containing the existing Girder token
+        """
+        self.token = token
 
     def getServerVersion(self, useCached=True):
         """
@@ -706,7 +715,8 @@ class GirderClient(object):
         }
         return self.createResource('collection', params)
 
-    def createFolder(self, parentId, name, description='', parentType='folder', public=None):
+    def createFolder(self, parentId, name, description='', parentType='folder',
+                     public=None, reuseExisting=False):
         """
         Creates and returns a folder.
 
@@ -716,7 +726,8 @@ class GirderClient(object):
             'parentId': parentId,
             'parentType': parentType,
             'name': name,
-            'description': description
+            'description': description,
+            'reuseExisting': reuseExisting
         }
         if public is not None:
             params['public'] = public
@@ -872,6 +883,82 @@ class GirderClient(object):
         with open(filepath, 'rb') as f:
             return self._uploadContents(obj, f, filesize, progressCallback=progressCallback)
 
+    def uploadStreamToFolder(self, folderId, stream, filename, size, reference=None, mimeType=None,
+                             progressCallback=None):
+        """
+        Uploads a file-like object to a folder, creating a new item in the process. If
+        the file has 0 bytes, no uploading will be performed, and no item will
+        be created.
+
+        :param folderId: ID of parent folder for file.
+        :param stream: Readable stream object.
+        :param filename: Filename used for Girder only.
+        :param size: The length of the file. This must be exactly equal to the
+            total number of bytes that will be read from ``stream``, otherwise
+            the upload will fail.
+        :param reference: optional reference to send along with the upload.
+        :param mimeType: MIME type for the file.
+        :param progressCallback: If passed, will be called after each chunk
+            with progress information. It passes a single positional argument
+            to the callable which is a dict of information about progress.
+        """
+        params = {
+            'parentType': 'folder',
+            'parentId': folderId,
+            'name': filename,
+            'size': size,
+            'mimeType': mimeType
+        }
+
+        if reference:
+            params['reference'] = reference
+
+        obj = self.post('file', params)
+
+        if '_id' not in obj:
+            raise Exception(
+                'After creating an upload token for a new file, expected '
+                'an object with an id. Got instead: ' + json.dumps(obj))
+
+        return self._uploadContents(obj, stream, size, progressCallback=progressCallback)
+
+    def uploadFileToFolder(self, folderId, filepath, reference=None, mimeType=None, filename=None,
+                           progressCallback=None):
+        """
+        Uploads a file to a folder, creating a new item in the process.  If
+        the file has 0 bytes, no uploading will be performed, and no item will
+        be created.
+
+        :param folderId: ID of parent folder for file.
+        :param filepath: path to file on disk.
+        :param reference: optional reference to send along with the upload.
+        :type reference: str
+        :param mimeType: MIME type for the file. Will be guessed if not passed.
+        :type mimeType: str or None
+        :param filename: path with filename used in Girder. Defaults to basename of filepath.
+        :param progressCallback: If passed, will be called after each chunk
+            with progress information. It passes a single positional argument
+            to the callable which is a dict of information about progress.
+        :type progressCallback: callable
+        :returns: the file that was created.
+        """
+        if filename is None:
+            filename = filepath
+        filename = os.path.basename(filename)
+        filepath = os.path.abspath(filepath)
+        filesize = os.path.getsize(filepath)
+
+        if filesize == 0:
+            return
+
+        if mimeType is None:
+            # Attempt to guess MIME type if not passed explicitly
+            mimeType, _ = mimetypes.guess_type(filepath)
+
+        with open(filepath, 'rb') as f:
+            return self.uploadStreamToFolder(folderId, f, filename, filesize, reference, mimeType,
+                                             progressCallback)
+
     def _uploadContents(self, uploadObj, stream, size, progressCallback=None):
         """
         Uploads contents of a file.
@@ -892,7 +979,7 @@ class GirderClient(object):
         offset = 0
         uploadId = uploadObj['_id']
 
-        with self.progressReporterCls(label=uploadObj['name'], length=size) as reporter:
+        with self.progressReporterCls(label=uploadObj.get('name', ''), length=size) as reporter:
 
             while True:
                 chunk = stream.read(min(self.MAX_CHUNK_SIZE, (size - offset)))
@@ -1031,7 +1118,7 @@ class GirderClient(object):
 
     def addMetadataToFolder(self, folderId, metadata):
         """
-        Takes an folder ID and a dictionary containing the metadata
+        Takes a folder ID and a dictionary containing the metadata
 
         :param folderId: ID of the folder to set metadata on.
         :param metadata: dictionary of metadata to set on folder.
@@ -1090,13 +1177,15 @@ class GirderClient(object):
         progressFileName = fileId
         if isinstance(path, six.string_types):
             progressFileName = os.path.basename(path)
-        req = self._requestFunc('get')(
-            '%sfile/%s/download' % (self.urlBase, fileId),
-            stream=True, headers={'Girder-Token': self.token})
+
+        url = '%sfile/%s/download' % (self.urlBase, fileId)
+        req = self._requestFunc('get')(url, stream=True, headers={'Girder-Token': self.token})
+        if not req.ok:
+            raise HttpError(req.status_code, req.text, url, 'GET')
         with tempfile.NamedTemporaryFile(delete=False) as tmp:
             with self.progressReporterCls(
                     label=progressFileName,
-                    length=int(req.headers['content-length'])) as reporter:
+                    length=int(req.headers.get('content-length', 0))) as reporter:
                 for chunk in req.iter_content(chunk_size=REQ_BUFFER_SIZE):
                     reporter.update(len(chunk))
                     tmp.write(chunk)
@@ -1392,34 +1481,37 @@ class GirderClient(object):
 
         return item
 
-    def _uploadFileToItem(self, localFile, parentItemId, filePath):
-        """Helper function to upload a file to an item
-        :param localFile: name of local file to upload
-        :param parentItemId: id of parent item in Girder to add file to
-        :param filePath: full path to the file
-        """
-        self.uploadFileToItem(parentItemId, filePath, filename=localFile)
-
-    def _uploadAsItem(self, localFile, parentFolderId, filePath, reuseExisting=False, dryRun=False):
+    def _uploadAsItem(self, localFile, parentFolderId, filePath, reuseExisting=False, dryRun=False,
+                      reference=None):
         """Function for doing an upload of a file as an item.
         :param localFile: name of local file to upload
         :param parentFolderId: id of parent folder in Girder
         :param filePath: full path to the file
         :param reuseExisting: boolean indicating whether to accept an existing item
             of the same name in the same location, or create a new one instead
+        :param reference: Option reference to send along with the upload.
         """
         if not self.progressReporterCls.reportProgress:
             print('Uploading Item from %s' % localFile)
         if not dryRun:
-            currentItem = self.loadOrCreateItem(
-                os.path.basename(localFile), parentFolderId, reuseExisting)
-            self._uploadFileToItem(localFile, currentItem['_id'], filePath)
-
-            for callback in self._itemUploadCallbacks:
-                callback(currentItem, filePath)
+            # If we are reusing existing items or have upload callbacks, then
+            # we need to know the item as part of the process.  If this is a
+            # zero-length file, we create an item.  Otherwise, we can just
+            # upload to the parent folder and never learn about the created
+            # item.
+            if reuseExisting or len(self._itemUploadCallbacks) or os.path.getsize(filePath) == 0:
+                currentItem = self.loadOrCreateItem(
+                    os.path.basename(localFile), parentFolderId, reuseExisting)
+                self.uploadFileToItem(
+                    currentItem['_id'], filePath, filename=localFile, reference=reference)
+                for callback in self._itemUploadCallbacks:
+                    callback(currentItem, filePath)
+            else:
+                self.uploadFileToFolder(
+                    parentFolderId, filePath, filename=localFile, reference=reference)
 
     def _uploadFolderAsItem(self, localFolder, parentFolderId, reuseExisting=False, blacklist=None,
-                            dryRun=False):
+                            dryRun=False, reference=None):
         """
         Take a folder and use its base name as the name of a new item. Then,
         upload its containing files into the new item as bitstreams.
@@ -1428,6 +1520,7 @@ class GirderClient(object):
         :param parentFolderId: Id of the destination folder for the new item.
         :param reuseExisting: boolean indicating whether to accept an existing item
             of the same name in the same location, or create a new one instead
+        :param reference: Option reference to send along with the upload.
         """
         blacklist = blacklist or []
         print('Creating Item from folder %s' % localFolder)
@@ -1447,14 +1540,14 @@ class GirderClient(object):
             print('Adding file %s, (%d of %d) to Item' % (currentFile, ind + 1, filecount))
 
             if not dryRun:
-                self._uploadFileToItem(currentFile, item['_id'], filepath)
+                self.uploadFileToItem(item['_id'], filepath, filename=currentFile)
 
         if not dryRun:
             for callback in self._itemUploadCallbacks:
                 callback(item, localFolder)
 
     def _uploadFolderRecursive(self, localFolder, parentId, parentType, leafFoldersAsItems=False,
-                               reuseExisting=False, blacklist=None, dryRun=False):
+                               reuseExisting=False, blacklist=None, dryRun=False, reference=None):
         """
         Function to recursively upload a folder and all of its descendants.
 
@@ -1465,6 +1558,7 @@ class GirderClient(object):
             files uploaded as single items
         :param reuseExisting: boolean indicating whether to accept an existing item
             of the same name in the same location, or create a new one instead
+        :param reference: Option reference to send along with the upload.
         """
         blacklist = blacklist or []
         if leafFoldersAsItems and self._hasOnlyFiles(localFolder):
@@ -1504,17 +1598,18 @@ class GirderClient(object):
                     # pass that as the parent_type
                     self._uploadFolderRecursive(
                         fullEntry, folder['_id'], 'folder', leafFoldersAsItems, reuseExisting,
-                        dryRun=dryRun)
+                        dryRun=dryRun, reference=reference)
                 else:
                     self._uploadAsItem(
-                        entry, folder['_id'], fullEntry, reuseExisting, dryRun=dryRun)
+                        entry, folder['_id'], fullEntry, reuseExisting, dryRun=dryRun,
+                        reference=reference)
 
             if not dryRun:
                 for callback in self._folderUploadCallbacks:
                     callback(folder, localFolder)
 
     def upload(self, filePattern, parentId, parentType='folder', leafFoldersAsItems=False,
-               reuseExisting=False, blacklist=None, dryRun=False):
+               reuseExisting=False, blacklist=None, dryRun=False, reference=None):
         """
         Upload a pattern of files.
 
@@ -1522,7 +1617,8 @@ class GirderClient(object):
         create a hierarchy on the server under the parentId.
 
         :param filePattern: a glob pattern for files that will be uploaded,
-            recursively copying any file folder structures.
+            recursively copying any file folder structures.  If this is a list
+            or tuple each item in it will be used in turn.
         :type filePattern: str
         :param parentId: Id of the parent in Girder or resource path.
         :type parentId: ObjectId or Unix-style path to the resource in Girder.
@@ -1537,33 +1633,37 @@ class GirderClient(object):
         :param dryRun: Set this to True to print out what actions would be taken, but
             do not actually communicate with the server.
         :type dryRun: bool
+        :param reference: Option reference to send along with the upload.
+        :type reference: str
         """
+        filePatternList = filePattern if isinstance(filePattern, (list, tuple)) else [filePattern]
         blacklist = blacklist or []
         empty = True
         parentId = self._checkResourcePath(parentId)
-        for currentFile in glob.iglob(filePattern):
-            empty = False
-            currentFile = os.path.normpath(currentFile)
-            filename = os.path.basename(currentFile)
-            if filename in blacklist:
-                if dryRun:
-                    print('Ignoring file %s as it is blacklisted' % filename)
-                continue
-            if os.path.isfile(currentFile):
-                if parentType != 'folder':
-                    raise Exception(
-                        'Attempting to upload an item under a %s. Items can only be added to '
-                        'folders.' % parentType)
+        for pattern in filePatternList:
+            for currentFile in glob.iglob(pattern):
+                empty = False
+                currentFile = os.path.normpath(currentFile)
+                filename = os.path.basename(currentFile)
+                if filename in blacklist:
+                    if dryRun:
+                        print('Ignoring file %s as it is blacklisted' % filename)
+                    continue
+                if os.path.isfile(currentFile):
+                    if parentType != 'folder':
+                        raise Exception(
+                            'Attempting to upload an item under a %s. Items can only be added to '
+                            'folders.' % parentType)
+                    else:
+                        self._uploadAsItem(
+                            os.path.basename(currentFile), parentId, currentFile, reuseExisting,
+                            dryRun=dryRun, reference=reference)
                 else:
-                    self._uploadAsItem(
-                        os.path.basename(currentFile), parentId, currentFile, reuseExisting,
-                        dryRun=dryRun)
-            else:
-                self._uploadFolderRecursive(
-                    currentFile, parentId, parentType, leafFoldersAsItems, reuseExisting,
-                    blacklist=blacklist, dryRun=dryRun)
+                    self._uploadFolderRecursive(
+                        currentFile, parentId, parentType, leafFoldersAsItems, reuseExisting,
+                        blacklist=blacklist, dryRun=dryRun, reference=reference)
         if empty:
-            print('No matching files: ' + filePattern)
+            print('No matching files: ' + repr(filePattern))
 
     def _checkResourcePath(self, objId):
         if isinstance(objId, six.string_types) and objId.startswith('/'):
